package hexIT;

import java.util.*;     // for Iterator

// Coord System & Example:
//                                          y-coord
//                                 0 1 2
//                                / / / 3
//                            0- . . . / 4
//                           1- . . . . /
//             x-coord      2- . . . . .       (has size = 3)
//                           3- . . . x    <-- that 'x' is at position (3,4)
//                            4- . . .
//
//
// notes:
//   - we need a (2*size-1) by (2*size-1) array to store all hexagons.
//     the representation of this hexagonal board quite naturally fits into a 2D array
//     - some of the array is wasted, but this it's worth the ease-of-use

public class HexagonalBoard extends Board implements java.io.Serializable, Cloneable
{
   private Hexagon[][]   tiles;     // coords stored as [x][y]
   private int sideLength;          // in tiles

   public static final int DIR_NW= 0,
                           DIR_NE= 1,
                           DIR_W= 2,
                           DIR_E= 3,
                           DIR_SW= 4,
                           DIR_SE= 5;
   
   public HexagonalBoard()
   {
   }
   
   public HexagonalBoard(int new_size)
   {
      super(new_size);

      sideLength= 2*size-1;
      tiles= new Hexagon[sideLength][sideLength];

      // create the empty tiles
      for (int i= 0; i < sideLength; i++)
         for (int j= 0; j < sideLength; j++)
            tiles[i][j]= new Hexagon(100, i, j);
   }

   // calculates the minimum # of hexes need to travel from one to the other.
   public int distance(Hexagon h1, Hexagon h2)
   {
      // this method may only work for dist <= 3 (haven't tested higher than that)
      int deltaX= h2.getHexX()-h1.getHexX();
      int deltaY= h2.getHexY()-h1.getHexY();
      
      if (   (sign(deltaX) == -1 && sign(deltaY) == 1) 
          || (sign(deltaX) == 1 && sign(deltaY) == -1) )      
         return (abs(deltaX) + abs(deltaY));
      
      return max(abs(deltaX),abs(deltaY));
   }
  
   public boolean isValidCoord(int x, int y)
   {
      int median= (sideLength/2);
      
      if (x < 0 || x >= sideLength || y < 0 || y >= sideLength)
         return false;     
      else if (x > median && x > y && (x-y) > median)         // bottom left region
         return false;      
      else if (y > median && y > x && (y-x) > median)         // top right region
         return false;
      
      return true;
   }
  
   public Hexagon getHexagon(int x, int y)
   {      
      return tiles[x][y];
   }
   
   public int getSideLength()
   {
      return sideLength;
   }
   
   public void moveObject(int x1, int y1, int x2, int y2)
   {
      if (!isValidCoord(x1, y1))
      {
         System.out.println("Invalid coordinates: (x1, y1) = (" + x1 + "," + y1 + ") in HexagonalBoard.moveObject()");
         return;
      }
      else if (!isValidCoord(x2, y2))
      {
         System.out.println("Invalid coordinates: (x2, y2) = (" + x2 + "," + y2 + ") in HexagonalBoard.moveObject()");
         return;
      }         
      else if (getHexagon(x1, y1).getObj() == null)
      {
         System.out.println("No object on source hexagon in HexagonalBoard.moveObject()");
         return;
      }
      else if (getHexagon(x2, y2).getObj() != null)
      {
         System.out.println("Object(collision) on destination hexagon in HexagonalBoard.moveObject()");
         return;
      }
         
      Drawable obj= getHexagon(x1, y1).getObj();
      getHexagon(x2, y2).putObj(obj);
      getHexagon(x1, y1).removeObj();      
   }
   
   public void moveObject(int x, int y, int direction)
   {
      switch(direction)
      {
         case HexagonalBoard.DIR_NW:
         {
            moveObject(x, y, x-1, y-1);
            break;
         }
         case HexagonalBoard.DIR_NE:
         {
            moveObject(x, y, x-1, y);
            break;
         }
         case HexagonalBoard.DIR_W:
         {
            moveObject(x, y, x, y-1);
            break;
         }
         case HexagonalBoard.DIR_E:
         {
            moveObject(x, y, x, y+1);
            break;
         }
         case HexagonalBoard.DIR_SW:
         {
            moveObject(x, y, x+1, y);
            break;
         }
         case HexagonalBoard.DIR_SE:
         {
            moveObject(x, y, x+1, y+1);
            break;
         }
         default: break;
      }         
   }

   public void putObject(int x, int y, Drawable obj)
   {
      getHexagon(x,y).putObj(obj);
   }

   public void removeObject(int x, int y)
   {
      getHexagon(x,y).removeObj();
   }

   public Iterator tilesIterator()
   {
      return new TileIterator(this);
   }
   
   public Hexagon getHexagonByPixel(int pixx, int pixy)
   {      
      Iterator i= tilesIterator();
      
      while (i.hasNext())
      {
         Hexagon h= (Hexagon)i.next();
         if (h.surroundsPixel(pixx, pixy))
            return h;
      }
      
      return null;
   }
   
   
   public List getFOV(int hexx, int hexy, int radius)
   {
      List hexes= new LinkedList();
      
      if (radius < 0)
         return hexes;
      
      hexes.add(getHexagon(hexx, hexy));
      getFOV_rec(hexx, hexy, radius, hexes);
      
      return hexes;
   }   
 
	// If there is a straight line of heaxgons (ie. you can get there by going
	// in one constant direction) between the source and destination, the the
	// direction (from source to dest) is returned. Otherwise, -1 is returned
	public int isStraightPath(int hexx1, int hexy1, int hexx2, int hexy2)
	{
		int d_hexx= (hexx2 - hexx1);
		int d_hexy= (hexy2 - hexy1);

		if (d_hexx == 0 && d_hexy != 0)
		{
			if (d_hexy > 0)
				return HexagonalBoard.DIR_E;
			else if (d_hexy < 0)
				return HexagonalBoard.DIR_W;
		}
		else if (d_hexx != 0 && d_hexy == 0)
		{
			if (d_hexx > 0)
				return HexagonalBoard.DIR_SW;
			else if (d_hexx < 0)
				return HexagonalBoard.DIR_NE;
		}
		else if (d_hexx != 0 && d_hexy != 0 && d_hexx == d_hexy)
		{
			if (d_hexx < 0)
				return HexagonalBoard.DIR_NW;
			else if (d_hexx > 0)
				return HexagonalBoard.DIR_SE;
		}

		return -1;
	}
		
   public List getLOS(int hexx1, int hexy1, int hexx2, int hexy2)
   {
		// to be implemented
      return null;
   }
      
   private void getFOV_rec(int hexx, int hexy, int radius, List curList)
   {
      if (radius == 0)
      {
         return;
      }      
      else
      {
         LinkedList oldList= new LinkedList(curList);
         
         for (int i= 0; i < oldList.size(); i++)
         {
            Hexagon h= (Hexagon)oldList.get(i);
            
            int cur_hexx= h.getHexX();
            int cur_hexy= h.getHexY();
            
            // left 1
            if (isValidCoord(cur_hexx, cur_hexy-1))           
               if (!oldList.contains(getHexagon(cur_hexx, cur_hexy-1)))
                  curList.add(getHexagon(cur_hexx, cur_hexy-1));

            // right 1
            if (isValidCoord(cur_hexx, cur_hexy+1))           
               if (!oldList.contains(getHexagon(cur_hexx, cur_hexy+1)))
                  curList.add(getHexagon(cur_hexx, cur_hexy+1));
            
            // northeast 1
            if (isValidCoord(cur_hexx-1, cur_hexy))
               if (!oldList.contains(getHexagon(cur_hexx-1, cur_hexy)))
                  curList.add(getHexagon(cur_hexx-1, cur_hexy));
            
            // northwest 1
            if (isValidCoord(cur_hexx-1, cur_hexy-1))           
               if (!oldList.contains(getHexagon(cur_hexx-1, cur_hexy-1)))
                  curList.add(getHexagon(cur_hexx-1, cur_hexy-1));
            
            // southwest 1
            if (isValidCoord(cur_hexx+1, cur_hexy))           
               if (!oldList.contains(getHexagon(cur_hexx+1, cur_hexy)))
                  curList.add(getHexagon(cur_hexx+1, cur_hexy));
               
            // southeast 1
            if (isValidCoord(cur_hexx+1, cur_hexy+1))           
               if (!oldList.contains(getHexagon(cur_hexx+1, cur_hexy+1)))
                  curList.add(getHexagon(cur_hexx+1, cur_hexy+1));
         }
         
         getFOV_rec(hexx, hexy, radius-1, curList);         
      }
   }
      
   // ------------------------------------
   
   public class TileIterator implements Iterator
   {
      private int x, y;
      HexagonalBoard b;
      
      public TileIterator(HexagonalBoard new_b)
      {
         b= new_b;
         x= y= 0;
         
         // go to the first valid coord
         for (int i= 0; i < sideLength; i++)
            for (int j= 0; j < sideLength; j++)
               if (b.isValidCoord(i,j))
               {
                  x= i;
                  y= j;
                  return;
               }
      }
      
      public boolean hasNext()
      {
         if (b.isValidCoord(x,y))         
            return true;
         else 
            return false;
         
      }
      
      public Object next()
      {
         int old_x= x, old_y= y;
         boolean done= false;
         boolean firstiter= true;
       
         if (!hasNext())
            return null;

         int i= 0, j= 0;
         
         for (i= x ; i < sideLength; i++)
         {                       
            for (j= (firstiter ? (y+1) : 0); j < sideLength; j++)
            {
               if (b.isValidCoord(i,j))
               {
                  done= true;
                  break;
               }
            }   

            firstiter= false;
            
            if (done)
               break;
            
         }
         
         x= i;
         y= j;
            
         return tiles[old_x][old_y];
      }
      
      public void remove()
      {
      }
   }
      
}
